// <!!-*-C++-*- file: binstream.cc --->
// <!!------------------------------------------------------------------------->
// <!! Copyright (C) 1996 Dietmar Kuehl >
// <!!   Universitaet Konstanz, Lehrstuhl fuer praktische Informatik I >
// <!!>
// <!! Permission to use, copy, modify, distribute and sell this >
// <!! software for any purpose is hereby granted without fee, provided >
// <!! that the above copyright notice appears in all copies and that >
// <!! both that copyright notice and this permission notice appear in >
// <!! supporting documentation. Dietmar Kuehl makes no representations
// <!! about the suitability of this software for any purpose. It is
// <!! provided "as is" without express or implied warranty.
// <!!-------------------------------------------------------------------------->

// Author: Dietmar Khl dietmar.kuehl@@uni-konstanz.de www.informatik.uni-konstanz.de/~kuehl
// Title:  Implementation of the binstream 

//------------------------------------------------------------------------------

#include <fstream>
#include <iostream>
#include <cassert>

#include "binstream.h"
#include "xdrstreambuf.h"

//------------------------------------------------------------------------------

#define CONST_CAST(type,expr)       const_cast<type>(expr)
#define STATIC_CAST(type,expr)      static_cast<type>(expr)
#define REINTERPRET_CAST(type,expr) reinterpret_cast<type>(expr)

//------------------------------------------------------------------------------
// Guarantee: The first argument to the following ctor may be '0'.
// However, in this case the corresponding 'streambuf' has to be set
// prior to an attempt to use the corresponding 'binios' object!

binios::binios(std::streambuf *sb, binios::operation op):
  sbuf(sb),
  xdrs(STATIC_CAST(XDR*, operator new(sizeof(XDR)))),
  len(0)
{
  xdrstreambuf_create(xdrs, sb, op == encode? XDR_ENCODE: XDR_DECODE);
}

binios::~binios()
{
  xdr_destroy(xdrs);
  operator delete(xdrs);
}

// stick with the old misnomer 'rdbuf' :-)

void binios::rdbuf(std::streambuf *sb)
{
  sbuf = sb;
  xdrstreambuf_reseat(xdrs, sb);
}

size_t binios::width(size_t w)
{
  size_t rc = len;
  len = w;
  return rc;
}

//------------------------------------------------------------------------------

obinstream::obinstream(std::streambuf *sb):
  binios(sb, encode)
{
}
obinstream::~obinstream()
{
}

obinstream &obinstream::operator<< (bool &b)
{
  bool_t tbool = b? TRUE: FALSE;
  xdr_bool(xdrstream(), &tbool);
  return *this;
}
obinstream &obinstream::operator<< (char &c) // NOTE: uses 4 bytes for a character!
{
  xdr_char(xdrstream(), &c);
  return *this;
}
obinstream &obinstream::operator<< (unsigned char &c)  { xdr_u_char(xdrstream(), &c); return *this; }
obinstream &obinstream::operator<< (signed char &c)    { xdr_char(xdrstream(), REINTERPRET_CAST(char*, &c)); return *this; }

obinstream &obinstream::operator<< (double &d)         { xdr_double(xdrstream(), &d); return *this; }
obinstream &obinstream::operator<< (float &f)          { xdr_float(xdrstream(), &f); return *this; }

obinstream &obinstream::operator<< (int &i)            { xdr_int(xdrstream(), &i); return *this; }
obinstream &obinstream::operator<< (long &l)           { xdr_long(xdrstream(), &l); return *this; }
obinstream &obinstream::operator<< (short &s)          { xdr_short(xdrstream(), &s); return *this; }

obinstream &obinstream::operator<< (unsigned int &i)   { xdr_u_int(xdrstream(), &i); return *this; }
obinstream &obinstream::operator<< (unsigned long &l)  { xdr_u_long(xdrstream(), &l); return *this; }
obinstream &obinstream::operator<< (unsigned short &s) { xdr_u_short(xdrstream(), &s); return *this; }

obinstream &obinstream::operator<< (char const *str)
{
  // Should this function also take 'width()' into account...?

  size_t len = strlen(str) + 1; // save the termination character, too

  // the definition of 'xdr_bytes()' takes a 'char**' instead of a
  // 'char *const*' as it does both reading and writing. However, for
  // writing it is indeed 'const'.
  int state = xdr_bytes(xdrstream(), CONST_CAST(char**, &str), &len, len);
  std::cout << "state: " << state << std::endl;

  return *this;
}

//------------------------------------------------------------------------------

ibinstream::ibinstream(std::streambuf *sb):
  binios(sb, decode)
{
}
ibinstream::~ibinstream()
{
}

ibinstream &ibinstream::operator>> (bool &b)
{
  bool_t tbool;
  xdr_bool(xdrstream(), &tbool);
  b = tbool == TRUE? true: false;
  return *this;
}
ibinstream &ibinstream::operator>> (char &c) // NOTE: uses 4 bytes for a character!
{
  xdr_char(xdrstream(), &c);
  return *this;
}
ibinstream &ibinstream::operator>> (unsigned char &c)  { xdr_u_char(xdrstream(), &c); return *this; }
ibinstream &ibinstream::operator>> (signed char &c)    { xdr_char(xdrstream(), REINTERPRET_CAST(char*, &c)); return *this; }

ibinstream &ibinstream::operator>> (double &d)         { xdr_double(xdrstream(), &d); return *this; }
ibinstream &ibinstream::operator>> (float &f)          { xdr_float(xdrstream(), &f); return *this; }

ibinstream &ibinstream::operator>> (int &i)            { xdr_int(xdrstream(), &i); return *this; }
ibinstream &ibinstream::operator>> (long &l)           { xdr_long(xdrstream(), &l); return *this; }
ibinstream &ibinstream::operator>> (short &s)          { xdr_short(xdrstream(), &s); return *this; }

ibinstream &ibinstream::operator>> (unsigned int &i)   { xdr_u_int(xdrstream(), &i); return *this; }
ibinstream &ibinstream::operator>> (unsigned long &l)  { xdr_u_long(xdrstream(), &l); return *this; }
ibinstream &ibinstream::operator>> (unsigned short &s) { xdr_u_short(xdrstream(), &s); return *this; }

ibinstream &ibinstream::operator>> (char *str)
{
  assert(width() != 0);

  size_t len = width();
  // is 'len' taken into account by 'xdr_bytes'?
  xdr_bytes(xdrstream(), &str, &len, len);

  return *this;
}

//------------------------------------------------------------------------------

ofbinstream::ofbinstream(char const *fname):
  obinstream((new std::filebuf())->open(fname, std::ios::out | std::ios::binary))
{
}

ofbinstream::~ofbinstream()
{
  std::filebuf *fb = STATIC_CAST(std::filebuf*,rdbuf());
  rdbuf(0);
  delete fb;
}

ifbinstream::ifbinstream(char const *fname):
  ibinstream((new std::filebuf())->open(fname, std::ios::in | std::ios::binary))
{
}

ifbinstream::~ifbinstream()
{
  std::filebuf *fb = STATIC_CAST(std::filebuf*,rdbuf());
  rdbuf(0);
  delete fb;
}
