// <!!-*-C++-*- file: xdrstreambuf.cc --->
// <!!------------------------------------------------------------------------->
// <!! Copyright (C) 1996 Dietmar Kuehl >
// <!!   Universitaet Konstanz, Lehrstuhl fuer praktische Informatik I >
// <!!>
// <!! Permission to use, copy, modify, distribute and sell this >
// <!! software for any purpose is hereby granted without fee, provided >
// <!! that the above copyright notice appears in all copies and that >
// <!! both that copyright notice and this permission notice appear in >
// <!! supporting documentation. Dietmar Kuehl makes no representations
// <!! about the suitability of this software for any purpose. It is
// <!! provided "as is" without express or implied warranty.
// <!!-------------------------------------------------------------------------->

// Author: Dietmar Khl dietmar.kuehl@@uni-konstanz.de www.informatik.uni-konstanz.de/~kuehl
// Title:  Implementation of an XDR to streambuf binding

//------------------------------------------------------------------------------

#include "xdrstreambuf.h"
#include <ios>

// The current implementation in libg++ does not define the method
// 'pubsync()' for 'streambuf' but has no access restrictions on 'sync()
// => Use this function.

#define STATIC_CAST(type,expr) static_cast<type>(expr)
#define REINTERPRET_CAST(type,expr) reinterpret_cast<type>(expr)

// The following somewhat defeats the naming of 'long' but then...

#if LONG_BIT == 64
  typedef int  long_t;
#else
  typedef long long_t;
#endif

//------------------------------------------------------------------------------
// The following functions need to be declared as 'extern "C"' to make
// sure an appropriate calling convention is used: These functions will
// be called from a C program.  The function 'xdrstreambuf_create()'
// does NOT have this requirements, as is can only be called from a C++
// function ('streambuf' has to be declared there).


extern "C" bool_t streambuf_getlong(XDR*, long_t*);
extern "C" bool_t streambuf_putlong(XDR*, long_t*);
extern "C" bool_t streambuf_getbytes(XDR*, caddr_t, int);
extern "C" bool_t streambuf_putbytes(XDR*, caddr_t, int);
extern "C" u_int  streambuf_getpostn(XDR*);
extern "C" bool_t streambuf_setpostn(XDR*, u_int);
extern "C" long_t *streambuf_inline(XDR*, int);
extern "C" void   streambuf_destroy(XDR*);

extern "C" bool_t streambuf_control(XDR*, int, void*);

//------------------------------------------------------------------------------

static XDR::xdr_ops xdrsb_ops = {
  streambuf_getlong,
  streambuf_putlong,
  streambuf_getbytes,
  streambuf_putbytes,
  streambuf_getpostn,
  streambuf_setpostn,
  streambuf_inline,
  streambuf_destroy,
  streambuf_control,
};

//------------------------------------------------------------------------------
// "ctor"

void xdrstreambuf_create(XDR *xdrs, std::streambuf *sb, xdr_op op)
{
  xdrs->x_op      = op;
  xdrs->x_ops     = &xdrsb_ops;
  xdrs->x_public  = 0; // bring into a defined state
  xdrs->x_private = REINTERPRET_CAST(caddr_t, sb);
  xdrs->x_base    = 0; // what is this for?
  xdrs->x_handy   = 0; // not used
}

//------------------------------------------------------------------------------
// "dtor"

void   streambuf_destroy(XDR *xdrs)
{
  // like for 'iostream' the 'streambuf' is owned by someone else.
  // However, synchronization would be nice:

  std::streambuf *sb = REINTERPRET_CAST(std::streambuf*, xdrs->x_private);
  if (sb != 0)
    sb->pubsync();
}

//------------------------------------------------------------------------------
// A method to implement the 'rdbuf()' mechanism of 'binios'

void xdrstreambuf_reseat(XDR *xdrs, std::streambuf *sb)
{
  REINTERPRET_CAST(std::streambuf*, xdrs->x_private)->pubsync();
  xdrs->x_private = REINTERPRET_CAST(caddr_t, sb);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// Why is 'x_getlong()' defined to work on 'long'? 'ntohl' and family works
// on 'unsigned long'!

bool_t streambuf_getlong(XDR *xdrs, long_t *ptr)
{
  std::streambuf *sb = REINTERPRET_CAST(std::streambuf*, xdrs->x_private);

  if (sb->sgetn(REINTERPRET_CAST(char*, ptr), 4) == 4)
  {
    *ptr = ntohl(*ptr);
    return TRUE;
  }
  else
    return FALSE;
}

bool_t streambuf_putlong(XDR *xdrs, long_t* ptr)
{
  std::streambuf *sb = REINTERPRET_CAST(std::streambuf*, xdrs->x_private);

  *ptr = htonl(*ptr);

  if (sb->sputn(REINTERPRET_CAST(char*, ptr), 4) == 4)
    return TRUE;
  else
    return FALSE;
}

//------------------------------------------------------------------------------

bool_t streambuf_getbytes(XDR *xdrs, caddr_t ptr, int len)
{
  std::streambuf *sb = REINTERPRET_CAST(std::streambuf*, xdrs->x_private);

  if (sb->sgetn(REINTERPRET_CAST(char*, ptr), len) == len)
    return TRUE;
  else
    return FALSE;
}

bool_t streambuf_putbytes(XDR *xdrs, caddr_t ptr, int len)
{
  std::streambuf *sb = REINTERPRET_CAST(std::streambuf*, xdrs->x_private);

  if (sb->sputn(REINTERPRET_CAST(char*, ptr), len) == len)
    return TRUE;
  else
    return FALSE;
}

//------------------------------------------------------------------------------

u_int  streambuf_getpostn(XDR *xdrs)
{
  std::streambuf *sb = REINTERPRET_CAST(std::streambuf*, xdrs->x_private);

  std::streampos pos = sb->pubseekoff(0, std::ios::cur, xdrs->x_op == XDR_ENCODE? std::ios::out: std::ios::in);
  return STATIC_CAST(u_int, pos);
}

bool_t streambuf_setpostn(XDR *xdrs, u_int p)
{
  std::streambuf     *sb = REINTERPRET_CAST(std::streambuf*, xdrs->x_private);
  std::streampos     pos = STATIC_CAST(std::streampos, p);
  std::ios::openmode which = xdrs->x_op == XDR_ENCODE? std::ios::out: std::ios::in;

  if (sb->pubseekpos(pos, which) != STATIC_CAST(std::streampos, std::streamoff(-1)))
    return TRUE;
  else
    return FALSE;
}

//------------------------------------------------------------------------------

long_t  *streambuf_inline(XDR * /* xdrs */, int /* len */)
{
  // To implement this, it would be necessary to gain specific access
  // to the 'streambuf' which cannot be guaranteed by all
  // 'streambuf's.  I'm not sure whether this method can usefully be
  // implemented... I hope, that it is at least not necessary to
  // implement this method to make the XDR conversions work!

  return 0;
}

//------------------------------------------------------------------------------

#include <iostream>

bool_t streambuf_control(XDR * /* xdrs */, int /* ??? */, void * /* ??? */)
{
  std::cout << "streambuf_control!" << std::endl;
  return FALSE;
}
